@@ -27,11 +27,14 @@
 #include <asm-generic/xenomai/wrappers.h> /* Read the generic portion. */
 #include <linux/interrupt.h>
 
+#define wrap_phys_mem_prot(filp,pfn,size,prot)	(prot)
+
 #define wrap_access_ok(task,addr,size) _access_ok((unsigned long)(addr),(size))
 
-#define rthal_irq_chip_enable(irq)   ({ rthal_irq_descp(irq)->chip->enable(irq); 0; })
-#define rthal_irq_chip_disable(irq)  ({ rthal_irq_descp(irq)->chip->disable(irq); 0; })
-#define rthal_irq_chip_end(irq)      ({ rthal_irq_descp(irq)->ipipe_end(irq, rthal_irq_descp(irq)); 0; })
+#define rthal_irq_desc_status(irq)	(rthal_irq_descp(irq)->status)
+#define rthal_irq_chip_enable(irq)	({ rthal_irq_descp(irq)->chip->enable(irq); 0; })
+#define rthal_irq_chip_disable(irq)	({ rthal_irq_descp(irq)->chip->disable(irq); 0; })
+#define rthal_irq_chip_end(irq)		({ rthal_irq_descp(irq)->ipipe_end(irq, rthal_irq_descp(irq)); 0; })
 
 typedef irq_handler_t rthal_irq_host_handler_t;
 
