@@ -4,7 +4,7 @@
  *
  * @note Copyright (C) 2005-2007 Jan Kiszka <jan.kiszka@web.de>
  * @note Copyright (C) 2005 Joerg Langenberg <joerg.langenberg@gmx.net>
- * @note Copyright (C) 2008 Gilles Chanteperdrix <gilles.chanteperdrix@gmail.com>
+ * @note Copyright (C) 2008 Gilles Chanteperdrix <gilles.chanteperdrix@xenomai.org>
  *
  * Xenomai is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by
@@ -1075,13 +1075,13 @@ void __rtdm_synch_flush(xnsynch_t *synch, unsigned long reason);
 
 static inline void rtdm_event_pulse(rtdm_event_t *event)
 {
-	trace_mark(xn_rtdm_event_pulse, "event %p", event);
+	trace_mark(xn_rtdm, event_pulse, "event %p", event);
 	__rtdm_synch_flush(&event->synch_base, 0);
 }
 
 static inline void rtdm_event_destroy(rtdm_event_t *event)
 {
-	trace_mark(xn_rtdm_event_destroy, "event %p", event);
+	trace_mark(xn_rtdm, event_destroy, "event %p", event);
 	__rtdm_synch_flush(&event->synch_base, XNRMID);
 	xnselect_destroy(&event->select_block);
 }
@@ -1110,7 +1110,7 @@ void rtdm_sem_up(rtdm_sem_t *sem);
 #ifndef DOXYGEN_CPP /* Avoid static inline tags for RTDM in doxygen */
 static inline void rtdm_sem_destroy(rtdm_sem_t *sem)
 {
-	trace_mark(xn_rtdm_sem_destroy, "sem %p", sem);
+	trace_mark(xn_rtdm, sem_destroy, "sem %p", sem);
 	__rtdm_synch_flush(&sem->synch_base, XNRMID);
 	xnselect_destroy(&sem->select_block);
 }
@@ -1132,7 +1132,7 @@ static inline void rtdm_mutex_unlock(rtdm_mutex_t *mutex)
 {
 	XENO_ASSERT(RTDM, !xnpod_asynch_p(), return;);
 
-	trace_mark(xn_rtdm_mutex_unlock, "mutex %p", mutex);
+	trace_mark(xn_rtdm, mutex_unlock, "mutex %p", mutex);
 
 	if (unlikely(xnsynch_wakeup_one_sleeper(&mutex->synch_base) != NULL))
 		xnpod_schedule();
@@ -1140,7 +1140,7 @@ static inline void rtdm_mutex_unlock(rtdm_mutex_t *mutex)
 
 static inline void rtdm_mutex_destroy(rtdm_mutex_t *mutex)
 {
-	trace_mark(xn_rtdm_mutex_destroy, "mutex %p", mutex);
+	trace_mark(xn_rtdm, mutex_destroy, "mutex %p", mutex);
 
 	__rtdm_synch_flush(&mutex->synch_base, XNRMID);
 }
