@@ -85,7 +85,7 @@ struct rthal_calibration_data rthal_tunables;
 
 rthal_trap_handler_t rthal_trap_handler;
 
-int rthal_realtime_faults[RTHAL_NR_CPUS][RTHAL_NR_FAULTS];
+unsigned rthal_realtime_faults[RTHAL_NR_CPUS][RTHAL_NR_FAULTS];
 
 volatile int rthal_sync_op;
 
@@ -129,9 +129,8 @@ void rthal_critical_exit(unsigned long flags)
  * routine, aimed at replacing the one provided by Adeos. Only very
  * specific situations actually require to override the default Adeos
  * setting for this parameter, like having to acknowledge non-standard
- * PIC hardware. @a ackfn should return a non-zero value to indicate
- * that the interrupt has been properly acknowledged. If @a ackfn is
- * NULL, the default Adeos routine will be used instead.
+ * PIC hardware. If @a ackfn is NULL, the default Adeos routine will
+ * be used instead.
  *
  * @param cookie A user-defined opaque cookie the HAL will pass to the
  * interrupt handler as its sole argument.
@@ -299,7 +298,8 @@ int rthal_irq_release(unsigned irq)
 
 int rthal_irq_host_pend(unsigned irq)
 {
-    return rthal_propagate_irq(irq);
+    rthal_propagate_irq(irq);
+    return 0;
 }
 
 /**
@@ -662,7 +662,7 @@ static int faults_read_proc(char *page,
         p += sprintf(p, "\n%3d: ", trap);
 
         for_each_online_cpu(cpu) {
-            p += sprintf(p, "%12d", rthal_realtime_faults[cpu][trap]);
+            p += sprintf(p, "%12u", rthal_realtime_faults[cpu][trap]);
         }
 
         p += sprintf(p, "    (%s)", rthal_fault_labels[trap]);
@@ -1089,3 +1089,7 @@ EXPORT_SYMBOL(rthal_proc_root);
 EXPORT_SYMBOL(rthal_init);
 EXPORT_SYMBOL(rthal_exit);
 EXPORT_SYMBOL(__rthal_generic_full_divmod64);
+
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,27)
+EXPORT_SYMBOL_GPL(kill_proc_info);
+#endif
