@@ -200,7 +200,7 @@ int xnheap_init(xnheap_t *heap,
 	 * static_part + pmapsize rounded to the minimum alignment
 	 * size.
 	*/
-	hdrsize = xnheap_overhead(heapsize, pagesize);
+	hdrsize = xnheap_internal_overhead(heapsize, pagesize);
 
 	/* Compute the page shiftmask from the page size (i.e. log2 value). */
 	for (pageshift = 0, shiftsize = pagesize;
@@ -947,18 +947,97 @@ int xnheap_check_block(xnheap_t *heap, void *block)
 #include <linux/vmalloc.h>
 #include <linux/mm.h>
 
-static DECLARE_DEVCLASS(xnheap_class);
-
 static DEFINE_XNQUEUE(kheapq);	/* Shared heap queue. */
 
+static inline void *__alloc_and_reserve_heap(size_t size, int kmflags)
+{
+	unsigned long vaddr, vabase;
+	void *ptr;
+
+	/* Size must be page-aligned. */
+
+	if ((kmflags & ~XNHEAP_GFP_NONCACHED) == 0) {
+		if (kmflags == 0)
+			ptr = vmalloc(size);
+		else
+			ptr = __vmalloc(size,
+					GFP_KERNEL | __GFP_HIGHMEM,
+					pgprot_noncached(PAGE_KERNEL));
+		if (ptr == NULL)
+			return NULL;
+
+		vabase = (unsigned long)ptr;
+		for (vaddr = vabase; vaddr < vabase + size; vaddr += PAGE_SIZE)
+			SetPageReserved(vmalloc_to_page((void *)vaddr));
+	} else {
+		/*
+		 * Otherwise, we have been asked for some kmalloc()
+		 * space. Assume that we can wait to get the required memory.
+		 */
+		if (size <= KMALLOC_MAX_SIZE)
+			ptr = kmalloc(size, kmflags | GFP_KERNEL);
+		else
+			ptr = (void *)__get_free_pages(kmflags | GFP_KERNEL,
+						       get_order(size));
+		if (ptr == NULL)
+			return NULL;
+
+		vabase = (unsigned long)ptr;
+		for (vaddr = vabase; vaddr < vabase + size; vaddr += PAGE_SIZE)
+			SetPageReserved(virt_to_page(vaddr));
+	}
+
+	return ptr;
+}
+
+static void __unreserve_and_free_heap(void *ptr, size_t size, int kmflags)
+{
+	unsigned long vaddr, vabase;
+
+	/* Size must be page-aligned. */
+
+	vabase = (unsigned long)ptr;
+
+	if ((kmflags & ~XNHEAP_GFP_NONCACHED) == 0) {
+		for (vaddr = vabase; vaddr < vabase + size; vaddr += PAGE_SIZE)
+			ClearPageReserved(vmalloc_to_page((void *)vaddr));
+
+		vfree(ptr);
+	} else {
+		for (vaddr = vabase; vaddr < vabase + size; vaddr += PAGE_SIZE)
+			ClearPageReserved(virt_to_page(vaddr));
+
+		if (size <= KMALLOC_MAX_SIZE)
+			kfree(ptr);
+		else
+			free_pages((unsigned long)ptr, get_order(size));
+	}
+}
+
 static void xnheap_vmclose(struct vm_area_struct *vma)
 {
-	xnheap_t *heap = (xnheap_t *)vma->vm_private_data;
-	atomic_dec(&heap->archdep.numaps);
+	xnheap_t *heap = vma->vm_private_data;
+	spl_t s;
+
+	xnlock_get_irqsave(&nklock, s);
+
+	if (atomic_dec_and_test(&heap->archdep.numaps)) {
+		if (heap->archdep.release) {
+			removeq(&kheapq, &heap->link);
+			xnlock_put_irqrestore(&nklock, s);
+			__unreserve_and_free_heap(heap->archdep.heapbase,
+						  xnheap_extentsize(heap),
+						  heap->archdep.kmflags);
+			xnlock_get_irqsave(&nklock, s);
+			heap->archdep.release(heap);
+		}
+	}
+
+	xnlock_put_irqrestore(&nklock, s);
 }
 
 static struct vm_operations_struct xnheap_vmops = {
-      close:&xnheap_vmclose
+      .close = &xnheap_vmclose
 };
 
 static int xnheap_open(struct inode *inode, struct file *file)
@@ -967,28 +1046,14 @@ static int xnheap_open(struct inode *inode, struct file *file)
 	return 0;
 }
 
-static int xnheap_release(struct inode *inode, struct file *file)
-{
-	xnheap_t *heap = (xnheap_t *)file->private_data;
-
-	/* Careful: the ioctl() binding might have not been issued. */
-
-	if (heap != NULL)
-		atomic_dec(&heap->archdep.numaps);
-
-	return 0;
-}
-
 static inline xnheap_t *__validate_heap_addr(void *addr)
 {
 	xnholder_t *holder;
 
-	/* Not time critical and seldomly called, so O(N) is ok here. */
-
 	for (holder = getheadq(&kheapq); holder;
 	     holder = nextq(&kheapq, holder))
-		if (link2heap(holder) == (xnheap_t *)addr)
-			return (xnheap_t *)addr;
+		if (link2heap(holder) == addr)
+			return addr;
 
 	return NULL;
 }
@@ -1009,7 +1074,6 @@ static int xnheap_ioctl(struct inode *inode,
 		goto unlock_and_exit;
 	}
 
-	atomic_inc(&heap->archdep.numaps);	/* Paired with xnheap_release() */
 	file->private_data = heap;
 
       unlock_and_exit:
@@ -1019,44 +1083,6 @@ static int xnheap_ioctl(struct inode *inode,
 	return err;
 }
 
-#ifdef CONFIG_MMU
-
-unsigned long __va_to_kva(unsigned long va)
-{
-	pgd_t *pgd;
-	pmd_t *pmd;
-	pte_t *ptep, pte;
-	unsigned long kva = 0;
-
-	pgd = pgd_offset_k(va);	/* Page directory in kernel map. */
-
-	if (!pgd_none(*pgd) && !pgd_bad(*pgd)) {
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 11)
-		/* Page middle directory -- account for PAE. */
-		pmd = pmd_offset(pud_offset(pgd, va), va);
-#else
-		/* Page middle directory. */
-		pmd = pmd_offset(pgd, va);
-#endif
-
-		if (!pmd_none(*pmd)) {
-			ptep = pte_offset_kernel(pmd, va);	/* Page table entry. */
-			pte = *ptep;
-
-			if (pte_present(pte)) {	/* Valid? */
-				kva = (unsigned long)page_address(pte_page(pte));	/* Page address. */
-				kva |= (va & (PAGE_SIZE - 1));	/* Add offset within page. */
-			}
-		}
-	}
-
-	return kva;
-}
-
-EXPORT_SYMBOL(__va_to_kva);
-
-#endif /* CONFIG_MMU */
-
 static int xnheap_mmap(struct file *file, struct vm_area_struct *vma)
 {
 	unsigned long offset, size, vaddr;
@@ -1071,15 +1097,11 @@ static int xnheap_mmap(struct file *file, struct vm_area_struct *vma)
 		return -EINVAL;	/* COW unsupported. */
 
 	offset = vma->vm_pgoff << PAGE_SHIFT;
-
-	if (offset != 0)
-		return -ENXIO;	/* We must map the entire heap. */
-
 	size = vma->vm_end - vma->vm_start;
 	heap = (xnheap_t *)file->private_data;
 
-	if (size != xnheap_extentsize(heap))
-		return -ENXIO;	/* Doesn't match the heap size. */
+	if (offset + size > xnheap_extentsize(heap))
+		return -ENXIO;
 
 	if (countq(&heap->extents) > 1)
 		/* Cannot map multi-extent heaps, we need the memory
@@ -1089,11 +1111,14 @@ static int xnheap_mmap(struct file *file, struct vm_area_struct *vma)
 	vma->vm_ops = &xnheap_vmops;
 	vma->vm_private_data = file->private_data;
 
-	vaddr = (unsigned long)heap->archdep.heapbase;
+	vaddr = (unsigned long)heap->archdep.heapbase + offset;
 
-	if (!heap->archdep.kmflags) {
+	if ((heap->archdep.kmflags & ~XNHEAP_GFP_NONCACHED) == 0) {
 		unsigned long maddr = vma->vm_start;
 
+		if (heap->archdep.kmflags == XNHEAP_GFP_NONCACHED)
+			vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);
+
 		while (size > 0) {
 			if (xnarch_remap_vm_page(vma, maddr, vaddr))
 				return -EAGAIN;
@@ -1102,124 +1127,19 @@ static int xnheap_mmap(struct file *file, struct vm_area_struct *vma)
 			vaddr += PAGE_SIZE;
 			size -= PAGE_SIZE;
 		}
-	} else if (xnarch_remap_io_page_range(vma,
+	} else if (xnarch_remap_io_page_range(file,vma,
 					      vma->vm_start,
 					      virt_to_phys((void *)vaddr),
-					      size, PAGE_SHARED))
+					      size, vma->vm_page_prot))
 		return -EAGAIN;
 
+	xnarch_fault_range(vma);
+
 	atomic_inc(&heap->archdep.numaps);
 
 	return 0;
 }
 
-static struct file_operations xnheap_fops = {
-	.owner = THIS_MODULE,
-	.open = &xnheap_open,
-	.release = &xnheap_release,
-	.ioctl = &xnheap_ioctl,
-	.mmap = &xnheap_mmap
-};
-
-static struct miscdevice xnheap_dev = {
-	XNHEAP_DEV_MINOR, "rtheap", &xnheap_fops
-};
-
-int xnheap_mount(void)
-{
-	struct class_device *cldev;
-
-	xnheap_class = class_create(THIS_MODULE, "rtheap");
-
-	if (IS_ERR(xnheap_class)) {
-		xnlogerr("error creating rtheap class, err=%ld.\n",
-			 PTR_ERR(xnheap_class));
-		return -EBUSY;
-	}
-
-	cldev = wrap_class_device_create(xnheap_class, NULL,
-					 MKDEV(MISC_MAJOR, XNHEAP_DEV_MINOR),
-					 NULL, "rtheap");
-	if (IS_ERR(cldev)) {
-		xnlogerr
-		    ("can't add device class, major=%d, minor=%d, err=%ld\n",
-		     MISC_MAJOR, XNHEAP_DEV_MINOR, PTR_ERR(cldev));
-		class_destroy(xnheap_class);
-		return -EBUSY;
-	}
-
-	if (misc_register(&xnheap_dev) < 0)
-		return -EBUSY;
-
-	return 0;
-}
-
-void xnheap_umount(void)
-{
-	misc_deregister(&xnheap_dev);
-	class_device_destroy(xnheap_class, MKDEV(MISC_MAJOR, XNHEAP_DEV_MINOR));
-	class_destroy(xnheap_class);
-}
-
-static inline void *__alloc_and_reserve_heap(size_t size, int kmflags)
-{
-	unsigned long vaddr, vabase;
-	void *ptr;
-
-	/* Size must be page-aligned. */
-
-	if (!kmflags) {
-		ptr = vmalloc(size);
-
-		if (!ptr)
-			return NULL;
-
-		vabase = (unsigned long)ptr;
-
-		for (vaddr = vabase; vaddr < vabase + size; vaddr += PAGE_SIZE)
-			SetPageReserved(virt_to_page(__va_to_kva(vaddr)));
-	} else {
-		/*
-		 * Otherwise, we have been asked for some kmalloc()
-		 * space. Assume that we can wait to get the required memory.
-		 */
-
-		ptr = kmalloc(size, kmflags | GFP_KERNEL);
-
-		if (!ptr)
-			return NULL;
-
-		vabase = (unsigned long)ptr;
-
-		for (vaddr = vabase; vaddr < vabase + size; vaddr += PAGE_SIZE)
-			SetPageReserved(virt_to_page(vaddr));
-	}
-
-	return ptr;
-}
-
-static inline void __unreserve_and_free_heap(void *ptr, size_t size,
-					     int kmflags)
-{
-	unsigned long vaddr, vabase;
-
-	/* Size must be page-aligned. */
-
-	vabase = (unsigned long)ptr;
-
-	if (!kmflags) {
-		for (vaddr = vabase; vaddr < vabase + size; vaddr += PAGE_SIZE)
-			ClearPageReserved(virt_to_page(__va_to_kva(vaddr)));
-
-		vfree(ptr);
-	} else {
-		for (vaddr = vabase; vaddr < vabase + size; vaddr += PAGE_SIZE)
-			ClearPageReserved(virt_to_page(vaddr));
-
-		kfree(ptr);
-	}
-}
-
 int xnheap_init_mapped(xnheap_t *heap, u_long heapsize, int memflags)
 {
 	void *heapbase;
@@ -1228,13 +1148,16 @@ int xnheap_init_mapped(xnheap_t *heap, u_long heapsize, int memflags)
 
 	/* Caller must have accounted for internal overhead. */
 	heapsize = xnheap_align(heapsize, PAGE_SIZE);
-	heapbase = __alloc_and_reserve_heap(heapsize, memflags);
 
+	if ((memflags & XNHEAP_GFP_NONCACHED)
+	    && memflags != XNHEAP_GFP_NONCACHED)
+		return -EINVAL;
+
+	heapbase = __alloc_and_reserve_heap(heapsize, memflags);
 	if (!heapbase)
 		return -ENOMEM;
 
 	err = xnheap_init(heap, heapbase, heapsize, PAGE_SIZE);
-
 	if (err) {
 		__unreserve_and_free_heap(heapbase, heapsize, memflags);
 		return err;
@@ -1242,6 +1165,7 @@ int xnheap_init_mapped(xnheap_t *heap, u_long heapsize, int memflags)
 
 	heap->archdep.kmflags = memflags;
 	heap->archdep.heapbase = heapbase;
+	heap->archdep.release = NULL;
 
 	xnlock_get_irqsave(&nklock, s);
 	appendq(&kheapq, &heap->link);
@@ -1250,25 +1174,67 @@ int xnheap_init_mapped(xnheap_t *heap, u_long heapsize, int memflags)
 	return 0;
 }
 
-int xnheap_destroy_mapped(xnheap_t *heap)
+int xnheap_destroy_mapped(xnheap_t *heap, void (*release)(struct xnheap *heap),
+			  void __user *mapaddr)
 {
+	int ret = 0, ccheck;
+	unsigned long len;
 	spl_t s;
 
+	ccheck = mapaddr ? 1 : 0;
+
 	xnlock_get_irqsave(&nklock, s);
 
-	if (atomic_read(&heap->archdep.numaps) > 0) {
+	if (atomic_read(&heap->archdep.numaps) > ccheck) {
+		heap->archdep.release = release;
 		xnlock_put_irqrestore(&nklock, s);
 		return -EBUSY;
 	}
 
-	removeq(&kheapq, &heap->link);
-	/* From now on, this heap cannot be mapped anymore. */
-
+	removeq(&kheapq, &heap->link); /* Prevent further mapping. */
 	xnlock_put_irqrestore(&nklock, s);
 
-	__unreserve_and_free_heap(heap->archdep.heapbase,
-				  xnheap_extentsize(heap), heap->archdep.kmflags);
-	return 0;
+	len = xnheap_extentsize(heap);
+
+	if (mapaddr) {
+		down_write(&current->mm->mmap_sem);
+		heap->archdep.release = NULL;
+		ret = do_munmap(current->mm, (unsigned long)mapaddr, len);
+		up_write(&current->mm->mmap_sem);
+	}
+
+	if (ret == 0) {
+		__unreserve_and_free_heap(heap->archdep.heapbase, len,
+					  heap->archdep.kmflags);
+		if (release) {
+			xnlock_get_irqsave(&nklock, s);
+			release(heap);
+			xnlock_put_irqrestore(&nklock, s);
+		}
+	}
+
+	return ret;
+}
+
+static struct file_operations xnheap_fops = {
+	.owner = THIS_MODULE,
+	.open = &xnheap_open,
+	.ioctl = &xnheap_ioctl,
+	.mmap = &xnheap_mmap
+};
+
+static struct miscdevice xnheap_dev = {
+	XNHEAP_DEV_MINOR, "rtheap", &xnheap_fops
+};
+
+int xnheap_mount(void)
+{
+	return misc_register(&xnheap_dev);
+}
+
+void xnheap_umount(void)
+{
+	misc_deregister(&xnheap_dev);
 }
 
 EXPORT_SYMBOL(xnheap_init_mapped);
