@@ -6,7 +6,7 @@
  * Copyright (C) 2004 The RTAI project <http://www.rtai.org>
  * Copyright (C) 2004 The HYADES project <http://www.hyades-itea.org>
  * Copyright (C) 2005 The Xenomai project <http://www.xenomai.org>
- * Copyright (C) 2006 Gilles Chanteperdrix <gilles.chanteperdrix@laposte.net>
+ * Copyright (C) 2006 Gilles Chanteperdrix <gilles.chanteperdrix@xenomai.org>
  *
  * Xenomai is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published
@@ -39,7 +39,6 @@
 #include <linux/wait.h>
 #include <linux/init.h>
 #include <linux/kthread.h>
-#include <asm/semaphore.h>
 #include <asm/signal.h>
 #include <nucleus/pod.h>
 #include <nucleus/heap.h>
@@ -205,13 +204,12 @@ static void rpi_push(xnthread_t *thread, int cpu)
 	int prio;
 	spl_t s;
 
-	/* non-RT shadows and RT shadows which disabled RPI cause the
-	   root priority to be lowered to its base level. The purpose
-	   of the following code is to enqueue the just thread
-	   whenever it involves RPI, and determine which priority to
-	   pick next for the root thread (i.e. the highest among RPI
-	   enabled threads, or the base level if none exists). */
-
+	/*
+	 * The purpose of the following code is to enqueue the thread
+	 * whenever it involves RPI, and determine which priority to
+	 * pick next for the root thread (i.e. the highest among RPI
+	 * enabled threads, or the base level if none exists).
+	 */
 	if (likely(xnthread_user_task(thread)->policy == SCHED_FIFO &&
 		   !xnthread_test_state(thread, XNRPIOFF))) {
 		xnlock_get_irqsave(&rpislot->lock, s);
@@ -819,21 +817,20 @@ static inline void request_syscall_restart(xnthread_t *thread,
 					   struct pt_regs *regs,
 					   int sysflags)
 {
+	int notify = 0;
+
 	if (xnthread_test_info(thread, XNKICKED)) {
-		if (__xn_interrupted_p(regs))
+		if (__xn_interrupted_p(regs)) {
 			__xn_error_return(regs,
 					  (sysflags & __xn_exec_norestart) ?
 					  -ERESTARTNOHAND : -ERESTARTSYS);
+			notify = !xnthread_test_state(thread, XNDEBUG);
+		}
 
 		xnthread_clear_info(thread, XNKICKED);
 	}
 
-	/* Relaxing due to a fault will trigger a notification from the
-	   trap handler when applicable, so we don't otherwise notify upon
-	   signal receipt, since testing syscall return values for -EINTR
-	   is still possible to detect such situation. */
-
-	xnshadow_relax(0);
+	xnshadow_relax(notify);
 }
 
 static inline void set_linux_task_priority(struct task_struct *p, int prio)
@@ -887,7 +884,7 @@ static void lostage_handler(void *cookie)
 		struct task_struct *p = rq->req[reqnum].task;
 		rq->out = (reqnum + 1) & (LO_MAX_REQUESTS - 1);
 
-		trace_mark(xn_nucleus_lostage_work, "reqnum %d comm %s pid %d",
+		trace_mark(xn_nucleus, lostage_work, "reqnum %d comm %s pid %d",
 			   reqnum, p->comm, p->pid);
 
 		switch (rq->req[reqnum].type) {
@@ -1017,7 +1014,8 @@ static int gatekeeper_thread(void *data)
 		   by a signal before we have been able to process the
 		   pending request, just ignore the latter. */
 
-		if (xnthread_user_task(thread)->state == TASK_INTERRUPTIBLE) {
+		if ((xnthread_user_task(thread)->state & ~TASK_ATOMICSWITCH)
+		    == TASK_INTERRUPTIBLE) {
 			rpi_pop(thread);
 			xnlock_get_irqsave(&nklock, s);
 #ifdef CONFIG_SMP
@@ -1101,7 +1099,7 @@ redo:
 	   preemption and using the TASK_ATOMICSWITCH cumulative state
 	   provided by Adeos to Linux tasks. */
 
-	trace_mark(xn_nucleus_shadow_gohard,
+	trace_mark(xn_nucleus, shadow_gohard,
 		   "thread %p thread_name %s comm %s",
 		   thread, xnthread_name(thread), this_task->comm);
 
@@ -1150,7 +1148,7 @@ redo:
 	if (rpi_p(thread))
 		rpi_clear_remote(thread);
 
-	trace_mark(xn_nucleus_shadow_hardened, "thread %p thread_name %s",
+	trace_mark(xn_nucleus, shadow_hardened, "thread %p thread_name %s",
 		   thread, xnthread_name(thread));
 
 	return 0;
@@ -1196,7 +1194,7 @@ void xnshadow_relax(int notify)
 	   domain to the Linux domain.  This will cause the Linux task
 	   to resume using the register state of the shadow thread. */
 
-	trace_mark(xn_nucleus_shadow_gorelax, "thread %p thread_name %s",
+	trace_mark(xn_nucleus, shadow_gorelax, "thread %p thread_name %s",
 		  thread, xnthread_name(thread));
 
 	ishield_on(thread);
@@ -1242,7 +1240,7 @@ void xnshadow_relax(int notify)
 	/* "current" is now running into the Linux domain on behalf of the
 	   root thread. */
 
-	trace_mark(xn_nucleus_shadow_relaxed,
+	trace_mark(xn_nucleus, shadow_relaxed,
 		   "thread %p thread_name %s comm %s",
 		   thread, xnthread_name(thread), current->comm);
 }
@@ -1293,7 +1291,10 @@ void xnshadow_exit(void)
  * Xenomai resource remains attached to it.
  *
  * - -EINVAL is returned if the thread control block does not bear the
- * XNSHADOW bit, or if the thread has already been mapped.
+ * XNSHADOW bit.
+ *
+ * - -EBUSY is returned if either the current Linux task or the
+ * associated shadow thread is already involved in a shadow mapping.
  *
  * Environments:
  *
@@ -1314,8 +1315,8 @@ int xnshadow_map(xnthread_t *thread, xncompletion_t __user *u_completion)
 	if (!xnthread_test_state(thread, XNSHADOW))
 		return -EINVAL;
 
-	if (xnthread_test_state(thread, XNMAPPED))
-		return -EINVAL;
+	if (xnshadow_thread(current) || xnthread_test_state(thread, XNMAPPED))
+		return -EBUSY;
 
 #ifdef CONFIG_MMU
 	if (!(current->mm->def_flags & VM_LOCKED))
@@ -1339,7 +1340,7 @@ int xnshadow_map(xnthread_t *thread, xncompletion_t __user *u_completion)
 		}
 	}
 
-	trace_mark(xn_nucleus_shadow_map,
+	trace_mark(xn_nucleus, shadow_map,
 		   "thread %p thread_name %s pid %d priority %d",
 		   thread, xnthread_name(thread), current->pid,
 		   xnthread_base_priority(thread));
@@ -1398,7 +1399,7 @@ void xnshadow_unmap(xnthread_t *thread)
 	xnthread_clear_state(thread, XNMAPPED);
 	rpi_pop(thread);
 
-	trace_mark(xn_nucleus_shadow_unmap,
+	trace_mark(xn_nucleus, shadow_unmap,
 		   "thread %p thread_name %s pid %d",
 		   thread, xnthread_name(thread), p ? p->pid : -1);
 
@@ -1465,7 +1466,7 @@ void xnshadow_start(xnthread_t *thread)
 	/* A shadow always starts in relaxed mode. */
 	rpi_push(thread, xnsched_cpu(thread->sched));
 
-	trace_mark(xn_nucleus_shadow_start, "thread %p thread_name %s",
+	trace_mark(xn_nucleus, shadow_start, "thread %p thread_name %s",
 		   thread, xnthread_name(thread));
 	xnpod_resume_thread(thread, XNDORMANT);
 
@@ -1593,14 +1594,14 @@ static int xnshadow_sys_bind(struct task_struct *curr, struct pt_regs *regs)
 	if (!check_abi_revision(abirev))
 		return -ENOEXEC;
 
-	if (!cap_raised(current->cap_effective, CAP_SYS_NICE) &&
+	if (!capable(CAP_SYS_NICE) &&
 	    (xn_gid_arg == -1 || !in_group_p(xn_gid_arg)))
 		return -EPERM;
 
 	/* Raise capabilities for the caller in case they are lacking yet. */
-	cap_raise(current->cap_effective, CAP_SYS_NICE);
-	cap_raise(current->cap_effective, CAP_IPC_LOCK);
-	cap_raise(current->cap_effective, CAP_SYS_RAWIO);
+	wrap_raise_cap(CAP_SYS_NICE);
+	wrap_raise_cap(CAP_IPC_LOCK);
+	wrap_raise_cap(CAP_SYS_RAWIO);
 
 	xnlock_get_irqsave(&nklock, s);
 
@@ -1894,9 +1895,11 @@ static inline int do_hisyscall_event(unsigned event, unsigned domid, void *data)
 	if (!__xn_reg_mux_p(regs))
 		goto linux_syscall;
 
-	/* Executing Xenomai services requires CAP_SYS_NICE, except for
-	   __xn_sys_bind which does its own checks. */
-	if (unlikely(!cap_raised(p->cap_effective, CAP_SYS_NICE)) &&
+	/*
+	 * Executing Xenomai services requires CAP_SYS_NICE, except for
+	 * __xn_sys_bind which does its own checks.
+	 */
+	if (unlikely(!cap_raised(current_cap(), CAP_SYS_NICE)) &&
 	    __xn_reg_mux(regs) != __xn_mux_code(0, __xn_sys_bind)) {
 		__xn_error_return(regs, -EPERM);
 		return RTHAL_EVENT_STOP;
@@ -1905,7 +1908,7 @@ static inline int do_hisyscall_event(unsigned event, unsigned domid, void *data)
 	muxid = __xn_mux_id(regs);
 	muxop = __xn_mux_op(regs);
 
-	trace_mark(xn_nucleus_syscall_histage,
+	trace_mark(xn_nucleus, syscall_histage,
 		   "thread %p thread_name %s muxid %d muxop %d",
 		   thread, thread ? xnthread_name(thread) : NULL,
 		   muxid, muxop);
@@ -2083,7 +2086,7 @@ static inline int do_losyscall_event(unsigned event, unsigned domid, void *data)
 	muxid = __xn_mux_id(regs);
 	muxop = __xn_mux_op(regs);
 
-	trace_mark(xn_nucleus_syscall_lostage,
+	trace_mark(xn_nucleus, syscall_lostage,
 		   "thread %p thread_name %s muxid %d muxop %d",
 		   xnpod_active_p() ? xnpod_current_thread() : NULL,
 		   xnpod_active_p() ? xnthread_name(xnpod_current_thread()) : NULL,
@@ -2147,7 +2150,7 @@ static inline void do_taskexit_event(struct task_struct *p)
 	if (!thread)
 		return;
 
-	if (xnthread_test_info(thread, XNDEBUG))
+	if (xnthread_test_state(thread, XNDEBUG))
 		unlock_timers();
 
 	if (xnpod_shadow_p())
@@ -2166,7 +2169,7 @@ static inline void do_taskexit_event(struct task_struct *p)
 	xnpod_schedule();
 
 	xnshadow_dereference_skin(magic);
-	trace_mark(xn_nucleus_shadow_exit, "thread %p thread_name %s",
+	trace_mark(xn_nucleus, shadow_exit, "thread %p thread_name %s",
 		   thread, xnthread_name(thread));
 }
 
@@ -2194,7 +2197,7 @@ static inline void do_schedule_event(struct task_struct *next)
 		 * order to encompass both the NPTL and LinuxThreads
 		 * behaviours.
 		 */
-		if (xnthread_test_info(threadin, XNDEBUG)) {
+		if (xnthread_test_state(threadin, XNDEBUG)) {
 			if (signal_pending(next)) {
 				sigset_t pending;
 				/*
@@ -2211,7 +2214,7 @@ static inline void do_schedule_event(struct task_struct *next)
 					goto no_ptrace;
 			}
 
-			xnthread_clear_info(threadin, XNDEBUG);
+			xnthread_clear_state(threadin, XNDEBUG);
 			unlock_timers();
 		}
 
@@ -2255,12 +2258,12 @@ static inline void do_sigwake_event(struct task_struct *p)
 	xnthread_t *thread = xnshadow_thread(p);
 	spl_t s;
 
-	if (!thread)
+	if (thread == NULL)
 		return;
 
 	xnlock_get_irqsave(&nklock, s);
 
-	if ((p->ptrace & PT_PTRACED) && !xnthread_test_info(thread, XNDEBUG)) {
+	if ((p->ptrace & PT_PTRACED) && !xnthread_test_state(thread, XNDEBUG)) {
 		sigset_t pending;
 
 		/* We already own the siglock. */
@@ -2269,16 +2272,35 @@ static inline void do_sigwake_event(struct task_struct *p)
 		if (sigismember(&pending, SIGTRAP) ||
 		    sigismember(&pending, SIGSTOP)
 		    || sigismember(&pending, SIGINT)) {
-			xnthread_set_info(thread, XNDEBUG);
+			xnthread_set_state(thread, XNDEBUG);
 			lock_timers();
 		}
 	}
 
 	if (xnthread_test_state(thread, XNRELAX))
 		goto unlock_and_exit;
+	/*
+	 * If we are kicking a shadow thread in primary mode, make
+	 * sure Linux won't schedule in its mate under our feet as a
+	 * result of running signal_wake_up(). The Xenomai scheduler
+	 * must remain in control for now, until we explicitly relax
+	 * the shadow thread to allow for processing the pending
+	 * signals. Make sure we keep the additional state flags
+	 * unmodified so that we don't break any undergoing ptrace.
+	 */
+	set_task_nowakeup(p);
 
-	if (thread == thread->sched->runthread)
-		xnsched_set_resched(thread->sched);
+	/*
+	 * Tricky case: a ready thread does not actually run, but
+	 * nevertheless waits for the CPU in primary mode, so we have
+	 * to make sure that it will be notified of the pending break
+	 * condition as soon as it enters xnpod_suspend_thread() from
+	 * a blocking Xenomai syscall.
+	 */
+	if (xnthread_test_state(thread, XNREADY)) {
+		xnthread_set_info(thread, XNKICKED);
+		goto unlock_and_exit;
+	}
 
 	if (xnpod_unblock_thread(thread))
 		xnthread_set_info(thread, XNKICKED);
@@ -2288,19 +2310,12 @@ static inline void do_sigwake_event(struct task_struct *p)
 		xnthread_set_info(thread, XNKICKED|XNBREAK);
 	}
 
-	/* If we are kicking a shadow thread, make sure Linux won't
-	   schedule in its mate under our feet as a result of running
-	   signal_wake_up(). The Xenomai scheduler must remain in
-	   control for now, until we explicitly relax the shadow
-	   thread to allow for processing the pending signals. Make
-	   sure we keep the additional state flags unmodified so that
-	   we don't break any undergoing ptrace. */
+	if (xnthread_test_info(thread, XNKICKED)) {
+		xnsched_set_resched(thread->sched);
+		xnpod_schedule();
+	}
 
-	set_task_nowakeup(p);
-
-	xnpod_schedule();
-
-      unlock_and_exit:
+unlock_and_exit:
 
 	xnlock_put_irqrestore(&nklock, s);
 }
