@@ -202,8 +202,12 @@ static void __heap_flush_private(xnheap_t *heap,
  *
  * - H_DMA causes the block pool associated to the heap to be
  * allocated in physically contiguous memory, suitable for DMA
- * operations with I/O devices. A 128Kb limit exists for @a heapsize
- * when this flag is passed.
+ * operations with I/O devices. @a heapsize beyond 128KiB will
+ * be rounded up to a two expononent allocation.
+ *
+ * - H_NONCACHED causes the heap not to be cached. This is necessary on
+ * platforms such as ARM to share a heap between kernel and user-space.
+ * Note that this flag is not compatible with the H_DMA flag.
  *
  * @return 0 is returned upon success. Otherwise:
  *
@@ -260,7 +264,9 @@ int rt_heap_create(RT_HEAP *heap, const char *name, size_t heapsize, int mode)
 
 		err = xnheap_init_mapped(&heap->heap_base,
 					 heapsize,
-					 (mode & H_DMA) ? GFP_DMA : 0);
+					 ((mode & H_DMA) ? GFP_DMA : 0)
+					 | ((mode & H_NONCACHED) ?
+					    XNHEAP_GFP_NONCACHED : 0));
 		if (err)
 			return err;
 
@@ -326,6 +332,25 @@ int rt_heap_create(RT_HEAP *heap, const char *name, size_t heapsize, int mode)
 	return err;
 }
 
+static void __heap_post_release(struct xnheap *h) /* nklock held, IRQs off */
+{
+	RT_HEAP *heap = container_of(h, RT_HEAP, heap_base);
+
+	removeq(heap->rqueue, &heap->rlink);
+
+#ifdef CONFIG_XENO_OPT_REGISTRY
+	if (heap->handle)
+		xnregistry_remove(heap->handle);
+#endif /* CONFIG_XENO_OPT_REGISTRY */
+
+	if (xnsynch_destroy(&heap->synch_base) == XNSYNCH_RESCHED)
+		/*
+		 * Some task has been woken up as a result of the
+		 * deletion: reschedule now.
+		 */
+		xnpod_schedule();
+}
+
 /**
  * @fn int rt_heap_delete(RT_HEAP *heap)
  *
@@ -360,7 +385,7 @@ int rt_heap_create(RT_HEAP *heap, const char *name, size_t heapsize, int mode)
  * Rescheduling: possible.
  */
 
-int rt_heap_delete(RT_HEAP *heap)
+int rt_heap_delete_inner(RT_HEAP *heap, void __user *mapaddr)
 {
 	int err = 0;
 	spl_t s;
@@ -386,43 +411,38 @@ int rt_heap_delete(RT_HEAP *heap)
 
 	xnlock_put_irqrestore(&nklock, s);
 
-	/* The heap descriptor has been marked as deleted before we
-	   released the superlock thus preventing any sucessful subsequent
-	   calls of rt_heap_delete(), so now we can actually destroy
-	   it safely. */
+	/*
+	 * The heap descriptor has been marked as deleted before we
+	 * released the superlock thus preventing any sucessful
+	 * subsequent calls of rt_heap_delete(), so now we can
+	 * actually destroy it safely.
+	 */
 
 #ifdef CONFIG_XENO_OPT_PERVASIVE
 	if (heap->mode & H_MAPPABLE)
-		err = xnheap_destroy_mapped(&heap->heap_base);
+		err = xnheap_destroy_mapped(&heap->heap_base,
+					    __heap_post_release, mapaddr);
 	else
 #endif /* CONFIG_XENO_OPT_PERVASIVE */
 		err = xnheap_destroy(&heap->heap_base, &__heap_flush_private, NULL);
 
 	xnlock_get_irqsave(&nklock, s);
 
-	if (!err) {
-		removeq(heap->rqueue, &heap->rlink);
-
-#ifdef CONFIG_XENO_OPT_REGISTRY
-		if (heap->handle)
-			xnregistry_remove(heap->handle);
-#endif /* CONFIG_XENO_OPT_REGISTRY */
-
-		if (xnsynch_destroy(&heap->synch_base) == XNSYNCH_RESCHED)
-			/* Some task has been woken up as a result of
-			   the deletion: reschedule now. */
-			xnpod_schedule();
-	} else
-		/* Deletion failed, likely due to a busy state;
-		 * restore the magic word, to re-enable the
-		 * descriptor. */
+	if (err)
 		heap->magic = XENO_HEAP_MAGIC;
+	else if (!(heap->mode & H_MAPPABLE))
+		__heap_post_release(&heap->heap_base);
 
 	xnlock_put_irqrestore(&nklock, s);
 
 	return err;
 }
 
+int rt_heap_delete(RT_HEAP *heap)
+{
+	return rt_heap_delete_inner(heap, NULL);
+}
+
 /**
  * @fn int rt_heap_alloc(RT_HEAP *heap,size_t size,RTIME timeout,void **blockp)
  *
