@@ -4,7 +4,7 @@
  *
  * @note Copyright (C) 2005-2007 Jan Kiszka <jan.kiszka@web.de>
  * @note Copyright (C) 2005 Joerg Langenberg <joerg.langenberg@gmx.net>
- * @note Copyright (C) 2008 Gilles Chanteperdrix <gilles.chanteperdrix@gmail.com>
+ * @note Copyright (C) 2008 Gilles Chanteperdrix <gilles.chanteperdrix@xenomai.org>
  *
  * Xenomai is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by
@@ -410,7 +410,7 @@ void rtdm_task_join_nrt(rtdm_task_t *task, unsigned int poll_delay)
 
 	XENO_ASSERT(RTDM, xnpod_root_p(), return;);
 
-	trace_mark(xn_rtdm_task_joinnrt, "thread %p poll_delay %u",
+	trace_mark(xn_rtdm, task_joinnrt, "thread %p poll_delay %u",
 		   task, poll_delay);
 
 	xnlock_get_irqsave(&nklock, s);
@@ -750,7 +750,7 @@ void rtdm_event_init(rtdm_event_t *event, unsigned long pending)
 {
 	spl_t s;
 
-	trace_mark(xn_rtdm_event_init, "event %p pending %lu", event, pending);
+	trace_mark(xn_rtdm, event_init, "event %p pending %lu", event, pending);
 
 	/* Make atomic for re-initialisation support */
 	xnlock_get_irqsave(&nklock, s);
@@ -831,7 +831,7 @@ void rtdm_event_signal(rtdm_event_t *event)
 	int resched = 0;
 	spl_t s;
 
-	trace_mark(xn_rtdm_event_signal, "event %p", event);
+	trace_mark(xn_rtdm, event_signal, "event %p", event);
 
 	xnlock_get_irqsave(&nklock, s);
 
@@ -926,7 +926,7 @@ int rtdm_event_timedwait(rtdm_event_t *event, nanosecs_rel_t timeout,
 
 	XENO_ASSERT(RTDM, !xnpod_unblockable_p(), return -EPERM;);
 
-	trace_mark(xn_rtdm_event_timedwait,
+	trace_mark(xn_rtdm, event_timedwait,
 		   "event %p timeout %Lu timeout_seq %p timeout_seq_value %Lu",
 		   event, (long long)timeout, timeout_seq, (long long)(timeout_seq ? *timeout_seq : 0));
 
@@ -1000,7 +1000,7 @@ void rtdm_event_clear(rtdm_event_t *event)
 {
 	spl_t s;
 
-	trace_mark(xn_rtdm_event_clear, "event %p", event);
+	trace_mark(xn_rtdm, event_clear, "event %p", event);
 
 	xnlock_get_irqsave(&nklock, s);
 
@@ -1097,7 +1097,7 @@ void rtdm_sem_init(rtdm_sem_t *sem, unsigned long value)
 {
 	spl_t s;
 
-	trace_mark(xn_rtdm_sem_init, "sem %p value %lu", sem, value);
+	trace_mark(xn_rtdm, sem_init, "sem %p value %lu", sem, value);
 
 	/* Make atomic for re-initialisation support */
 	xnlock_get_irqsave(&nklock, s);
@@ -1211,7 +1211,7 @@ int rtdm_sem_timeddown(rtdm_sem_t *sem, nanosecs_rel_t timeout,
 
 	XENO_ASSERT(RTDM, !xnpod_unblockable_p(), return -EPERM;);
 
-	trace_mark(xn_rtdm_sem_timedwait,
+	trace_mark(xn_rtdm, sem_timedwait,
 		   "sem %p timeout %Lu timeout_seq %p timeout_seq_value %Lu",
 		   sem, (long long)timeout, timeout_seq, (long long)(timeout_seq ? *timeout_seq : 0));
 
@@ -1279,7 +1279,7 @@ void rtdm_sem_up(rtdm_sem_t *sem)
 {
 	spl_t s;
 
-	trace_mark(xn_rtdm_sem_up, "sem %p", sem);
+	trace_mark(xn_rtdm, sem_up, "sem %p", sem);
 
 	xnlock_get_irqsave(&nklock, s);
 
@@ -1501,7 +1501,7 @@ int rtdm_mutex_timedlock(rtdm_mutex_t *mutex, nanosecs_rel_t timeout,
 	spl_t s;
 	int err = 0;
 
-	trace_mark(xn_rtdm_mutex_timedlock,
+	trace_mark(xn_rtdm, mutex_timedlock,
 		   "mutex %p timeout %Lu timeout_seq %p timeout_seq_value %Lu",
 		   mutex, (long long)timeout, timeout_seq, (long long)(timeout_seq ? *timeout_seq : 0));
 
@@ -1774,6 +1774,7 @@ static int rtdm_mmap_buffer(struct file *filp, struct vm_area_struct *vma)
 {
 	struct rtdm_mmap_data *mmap_data = filp->private_data;
 	unsigned long vaddr, paddr, maddr, size;
+	int ret;
 
 	vma->vm_ops = mmap_data->vm_ops;
 	vma->vm_private_data = mmap_data->vm_private_data;
@@ -1803,15 +1804,21 @@ static int rtdm_mmap_buffer(struct file *filp, struct vm_area_struct *vma)
 			vaddr += PAGE_SIZE;
 			mapped_size += PAGE_SIZE;
 		}
-		return 0;
+		xnarch_fault_range(vma);
+		ret = 0;
 	} else
 #endif /* CONFIG_MMU */
 	if (mmap_data->src_paddr)
-		return xnarch_remap_io_page_range(vma, maddr, paddr,
-						  size, PAGE_SHARED);
-	else
-		return xnarch_remap_kmem_page_range(vma, maddr, paddr,
-						    size, PAGE_SHARED);
+	  	ret = xnarch_remap_io_page_range(filp, vma, maddr, paddr,
+						 size, PAGE_SHARED);
+	else {
+		ret = xnarch_remap_kmem_page_range(vma, maddr, paddr,
+						   size, PAGE_SHARED);
+		if (!ret)
+			xnarch_fault_range(vma);
+	}
+
+	return ret;
 }
 
 static struct file_operations rtdm_mmap_fops = {
