@@ -411,7 +411,7 @@ static int __wind_task_nametoid(struct task_struct *curr, struct pt_regs *regs)
 {
 	char name[XNOBJECT_NAME_LEN];
 	WIND_TCB_PLACEHOLDER ph;
-	TASK_ID task_id;
+	xnhandle_t handle;
 
 	if (!__xn_access_ok
 	    (curr, VERIFY_WRITE, __xn_reg_arg2(regs), sizeof(ph)))
@@ -429,12 +429,11 @@ static int __wind_task_nametoid(struct task_struct *curr, struct pt_regs *regs)
 			       sizeof(name) - 1);
 	name[sizeof(name) - 1] = '\0';
 
-	task_id = taskNameToId(name);
-
-	if (task_id == ERROR)
+	handle = taskNameToHandle(name);
+	if (handle == XN_NO_HANDLE)
 		return wind_errnoget();
 
-	ph.handle = task_id;	/* Copy back the task handle. */
+	ph.handle = handle; /* Copy back the task handle. */
 
 	__xn_copy_to_user(curr, (void __user *)__xn_reg_arg2(regs), &ph,
 			  sizeof(ph));
@@ -442,6 +441,26 @@ static int __wind_task_nametoid(struct task_struct *curr, struct pt_regs *regs)
 	return 0;
 }
 
+/*
+ * int __wind_task_setmode(int clrmask, int setmask, int *mode_r)
+ */
+
+static int __wind_task_setmode(struct task_struct *curr, struct pt_regs *regs)
+{
+	int setmask, clrmask, mode_r;
+
+	clrmask = __xn_reg_arg1(regs);
+	setmask = __xn_reg_arg2(regs);
+
+	/* Primary required: current thread must be valid. */
+	mode_r = xnpod_set_thread_mode(xnpod_current_thread(),
+				       clrmask, setmask);
+	if (__xn_reg_arg3(regs))
+		__xn_copy_to_user(curr, (void __user *)__xn_reg_arg3(regs),
+				  &mode_r, sizeof(mode_r));
+	return 0;
+}
+
 /*
  * int __wind_sem_bcreate(int flags, SEM_B_STATE state, SEM_ID *psem_id)
  */
@@ -709,15 +728,16 @@ static int __wind_taskinfo_get(struct task_struct *curr, struct pt_regs *regs)
 		return -EFAULT;
 
 	pTcb = (WIND_TCB *)xnregistry_fetch(handle);
-
 	if (!pTcb)
 		return S_objLib_OBJ_ID_ERROR;
 
 	err = taskInfoGet((TASK_ID)pTcb, &desc);
-
-	if (!err)
+	if (!err) {
+		/* Replace the kernel-based pointer by the userland handle. */
+		desc.td_tid = handle;
 		__xn_copy_to_user(curr, (void __user *)__xn_reg_arg2(regs), &desc,
 				  sizeof(desc));
+	}
 
 	return err;
 }
@@ -732,11 +752,12 @@ static int __wind_errno_taskset(struct task_struct *curr, struct pt_regs *regs)
 	int errcode = __xn_reg_arg2(regs);
 	WIND_TCB *pTcb;
 
-	if (handle)
-		pTcb = (WIND_TCB *)xnregistry_fetch(handle);
-	else
-		pTcb = __wind_task_current(curr);
+	if (!handle) {
+		wind_errnoset(errcode);
+		return 0;
+	}
 
+	pTcb = (WIND_TCB *)xnregistry_fetch(handle);
 	if (!pTcb)
 		return S_objLib_OBJ_ID_ERROR;
 
@@ -760,18 +781,17 @@ static int __wind_errno_taskget(struct task_struct *curr, struct pt_regs *regs)
 	    (curr, VERIFY_WRITE, __xn_reg_arg2(regs), sizeof(errcode)))
 		return -EFAULT;
 
-	if (handle)
+	if (!handle)
+		errcode = wind_errnoget();
+	else {
 		pTcb = (WIND_TCB *)xnregistry_fetch(handle);
-	else
-		pTcb = __wind_task_current(curr);
+		if (!pTcb)
+			return S_objLib_OBJ_ID_ERROR;
 
-	if (!pTcb)
-		return S_objLib_OBJ_ID_ERROR;
-
-	errcode = errnoOfTaskGet((TASK_ID) pTcb);
-
-	if (errcode == ERROR)
-		return wind_errnoget();
+		errcode = errnoOfTaskGet((TASK_ID) pTcb);
+		if (errcode == ERROR)
+			return wind_errnoget();
+	}
 
 	__xn_copy_to_user(curr, (void __user *)__xn_reg_arg2(regs), &errcode,
 			  sizeof(errcode));
@@ -1325,6 +1345,7 @@ static xnsysent_t __systab[] = {
 	[__vxworks_task_delay] = {&__wind_task_delay, __xn_exec_primary},
 	[__vxworks_task_verifyid] = {&__wind_task_verifyid, __xn_exec_any},
 	[__vxworks_task_nametoid] = {&__wind_task_nametoid, __xn_exec_any},
+	[__vxworks_task_setmode] = {&__wind_task_setmode, __xn_exec_primary},
 	[__vxworks_sem_bcreate] = {&__wind_sem_bcreate, __xn_exec_any},
 	[__vxworks_sem_ccreate] = {&__wind_sem_ccreate, __xn_exec_any},
 	[__vxworks_sem_mcreate] = {&__wind_sem_mcreate, __xn_exec_any},
